#############################
#    Identification of      #
#    dynamic systems        #
#                           #
#  author: Tomas Zimek      #
#    date: 24/5/2019        #
#############################

try:
    import wx
    import matplotlib.figure
    import matplotlib.pyplot
    import matplotlib.backends.backend_wxagg
    import numpy as np
    from scipy import signal
except ImportError as error:
    raise ImportError('The wxPython, numpy, control and matplotlib modules are required to run this program.')

import os, sys

width = 950
height = 650

class MatlabPlot(wx.Panel):
    def __init__(self, parent):
        wx.Panel.__init__(self, parent, style = wx.NO_FULL_REPAINT_ON_RESIZE)

        self.figure = matplotlib.figure.Figure()
        self.canvas = matplotlib.backends.backend_wxagg.FigureCanvasWxAgg(self, -1, self.figure)

        self.flag = False
        self.Bind(wx.EVT_IDLE, self.OnEvtIdle)
        self.Bind(wx.EVT_SIZE, self.OnEvtSize)

        self.SetToResize()
        self.DrawSelf()

    def OnEvtIdle(self, event):
        if self.flag:
            self.flag = False
            self.SetToResize()

    def OnEvtSize(self, event):
        self.flag = True

    def SetToResize(self):
        pixels = tuple(self.GetSize())
        self.SetSize(pixels)
        self.canvas.SetSize(pixels)
        self.figure.set_size_inches([float(x) / self.figure.get_dpi() for x in pixels])

    def DrawSelf(self):
        self.canvas.draw()

# END class MatlabPlot ----------------------------------------------------

class Identification(wx.Frame):
    def __init__(self, parent):
        wx.Frame.__init__(self, parent, title = 'Identification of dynamic systems', size = (width, height),
                                        style = wx.MINIMIZE_BOX | wx.MAXIMIZE_BOX | wx.SYSTEM_MENU | wx.CAPTION | wx.CLOSE_BOX | wx.RESIZE_BORDER ) # | wx.RESIZE_BORDER
        self.mainIconFile = 'icon.png'

        self.SetMinSize((width, height))
        self.InitProp()
        self.InitUI()
        self.Centre()
        self.Show()

    #----------------------------------------------------------------------
    def InitProp(self):
        self.IN = None
        self.OUT = None
        self.SAMPLE_PER = None
        self.TIME = None
        self.flagAnalyzed = False

        #LS method data
        self.Y = None
        self.PHI = None
        self.THETA = None
        self.LSmodel = None
        self.YLS = None
        self.LSOUT = None
        self.T1 = None

        #RLS method data
        self.THETA_R = None
        self.THETA_BUFF = None
        self.RLSmodel = None
        self.YRLS = None
        self.RLSOUT = None
        self.T2 = None

    #----------------------------------------------------------------------
    def InitUI(self):
        mainIcon = wx.Icon()
        mainIcon.CopyFromBitmap(wx.Bitmap(self.mainIconFile, wx.BITMAP_TYPE_ANY))
        self.SetIcon(mainIcon)

        font9 = wx.Font(9, wx.DEFAULT, wx.NORMAL , wx.NORMAL)
        font15B = wx.Font(15, wx.DEFAULT, wx.NORMAL , wx.BOLD)

        self.statusbar = self.CreateStatusBar(1)
        self.statusbar.SetStatusText('Ready')

        backgroundVBox = wx.BoxSizer(wx.VERTICAL)

        self.panel = wx.Panel(self, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize)

        vBox = wx.BoxSizer(wx.VERTICAL)
        #1
        hBox1 = wx.BoxSizer(wx.HORIZONTAL)
        st1 = wx.StaticText(self.panel,  wx.ID_ANY, 'Identification of dynamic systems')
        st1.SetFont(font15B)
        hBox1.Add(st1, 0, wx.ALL, 5)
        vBox.Add(wx.StaticLine(self.panel), 0, wx.ALL | wx.EXPAND, 5)
        vBox.Add(hBox1, 0, wx.CENTER)
        vBox.Add(wx.StaticLine(self.panel), 0, wx.ALL | wx.EXPAND, 5)

        #2
        hBox2 = wx.BoxSizer(wx.HORIZONTAL)
        st2 = wx.StaticText(self.panel,  wx.ID_ANY, 'Location of data:')
        st2.SetFont(font9)
        self.fpc2 = wx.FilePickerCtrl(self.panel, wx.ID_ANY,'', size = (525, -1), wildcard = 'CSV files (*.csv) |*.csv')
        self.btn2Load = wx.Button(self.panel, label='Load', size = (85, -1))
        self.btn3Analyze = wx.Button(self.panel, label='Analyze', size = (100, -1))
        self.btn3Analyze.Disable()
        self.Bind(wx.EVT_BUTTON, self.PressLoad, self.btn2Load)
        self.Bind(wx.EVT_BUTTON, self.PressAnalyze, self.btn3Analyze)
        hBox2.Add(st2, 0, wx.TOP | wx.LEFT, 15)
        hBox2.Add(self.fpc2, 0, wx.ALL | wx.EXPAND, 10)
        hBox2.Add(self.btn2Load, 0, wx.TOP | wx.BOTTOM, 10)
        hBox2.Add(self.btn3Analyze, 0, wx.LEFT | wx.TOP | wx.BOTTOM, 10)
        vBox.Add(hBox2)
        vBox.Add(wx.StaticLine(self.panel), 0, wx.ALL | wx.EXPAND, 5)

        #3 with Plotpanel
        hBox3 = wx.BoxSizer(wx.HORIZONTAL)
        self.panelPlot3 = MatlabPlot(self.panel)
        vBox3 = wx.BoxSizer(wx.VERTICAL)
        self.cb31 = wx.CheckBox(self.panel, label = 'Input signal')
        self.cb32 = wx.CheckBox(self.panel, label = 'Output signal')
        self.cb33 = wx.CheckBox(self.panel, label = 'Method of Least Squares')
        self.cb34 = wx.CheckBox(self.panel, label = 'Recursive m. of Least Squares')
        self.cb31.Disable()
        self.cb32.Disable()
        self.cb33.Disable()
        self.cb34.Disable()
        self.tc3 = wx.TextCtrl(self.panel, style = wx.TE_MULTILINE | wx.TE_READONLY)
        vBox3.Add(self.cb31, 0, wx.ALL, 2)
        vBox3.Add(self.cb32, 0, wx.ALL, 2)
        vBox3.Add(self.cb33, 0, wx.ALL, 2)
        vBox3.Add(self.cb34, 0, wx.ALL, 2)
        vBox3.Add(wx.StaticLine(self.panel), 0, wx.ALL | wx.EXPAND, 5)
        vBox3.Add(self.tc3, 1, wx.EXPAND)
        hBox3.Add(self.panelPlot3, 3, wx.ALL | wx.EXPAND, 5)
        hBox3.Add(vBox3, 1, wx.ALL | wx.EXPAND, 5)
        vBox.Add(hBox3, 1, wx.EXPAND)
        vBox.Add(wx.StaticLine(self.panel), 0, wx.ALL | wx.EXPAND, 5)
        # END vertical box inside panel

        self.panel.SetSizer(vBox, 1)
        self.panel.Layout()
        #vBox.Fit(self.panel)
        backgroundVBox.Add(self.panel, 1, wx.EXPAND, 5)

        self.SetSizer(backgroundVBox)
        self.Layout()

        self.Bind(wx.EVT_CHECKBOX, self.OnCheckboxEvent)

        sys.stdout = self.tc3 #redirrecting print function to GUI for debugging purposes also
        #self.panel.SetBackgroundColour((0xff, 0xff, 0xff))

    #----------------------------------------------------------------------
    def PressLoad(self, e):
        fileLocation = self.fpc2.GetPath()
        if not os.path.exists(fileLocation):
            text = 'File ' + os.path.basename(fileLocation) + ' does not exist.\nPlease check the path and try it again.'
            dlg = wx.MessageDialog(self, text, 'Select a file.', wx.OK | wx.ICON_EXCLAMATION)
            dlg.ShowModal()
            dlg.Destroy()
            self.statusbar.SetStatusText('Path to file does not exist.')
            return

        dlg = wx.TextEntryDialog(self, 'Sample period [sec] e.g.  0.01:')
        if not dlg.ShowModal() == wx.ID_OK:
            dlg.Destroy()
            return

        try:
            val = float(dlg.GetValue())
        except:
            dlg2 = wx.MessageDialog(self, 'Please enter valid number (e.g.  0.01)', 'Something is wrong.', wx.OK | wx.ICON_EXCLAMATION)
            dlg2.ShowModal()
            dlg2.Destroy()
            return

        self.SAMPLE_PER = val
        dlg.Destroy()

        loadedIO = np.genfromtxt(fileLocation, delimiter = ',')
        self.IN = loadedIO[:, 0]
        self.OUT = loadedIO[:, 1]
        self.TIME = np.arange(0, len(self.IN)*self.SAMPLE_PER, self.SAMPLE_PER)

        self.statusbar.SetStatusText('Data was successfully loaded.')
        self.tc3.Value = 'Data for processing was loaded.\n'
        self.tc3.AppendText('\nSample period = ' + str(self.SAMPLE_PER) + '.\n')
        self.PlotLoadedData()
        self.cb31.SetValue(True)
        self.cb32.SetValue(True)
        self.cb31.Enable()
        self.cb32.Enable()
        self.cb33.SetValue(False)
        self.cb34.SetValue(False)
        self.cb33.Disable()
        self.cb34.Disable()
        self.btn3Analyze.Enable()

    #----------------------------------------------------------------------
    def PressAnalyze(self, e):
        self.btn3Analyze.Disable()
        self.LSMethod()
        self.RLSMethod()
        self.CreateModelsAndOutputs() #ommit if plot will be implemented
        self.PrintResults()
        self.PlotResults()
        self.cb33.SetValue(True)
        self.cb34.SetValue(True)
        self.cb33.Enable()
        self.cb34.Enable()
        self.flagAnalyzed = True

    #----------------------------------------------------------------------
    def PlotLoadedData(self):
        self.panelPlot3.figure.clf()
        axes = self.panelPlot3.figure.gca()
        #axes.plot(self.TIME, self.IN, 'b', self.TIME, self.OUT, 'r')
        axes.plot(self.TIME, self.IN, 'b', linewidth=2)
        axes.plot(self.TIME, self.OUT, 'r', linewidth=2)
        axes.grid(True)
        axes.set_xlabel('t [s]')
        axes.set_ylabel('u(t), y(t)')
        axes.set_title('Input and output signals from identified system')
        axes.legend(['input u(t)', 'output y(t)'])
        self.panelPlot3.DrawSelf()

    #----------------------------------------------------------------------
    def LSMethod(self):
        self.Y = self.OUT[2:]
        self.PHI = np.column_stack([-1*self.OUT[1:-1], -1*self.OUT[:-2], self.IN[1:-1], self.IN[:-2]])
        self.THETA = (np.linalg.inv(self.PHI.T @ self.PHI)) @ (self.PHI.T @ self.Y)

    #----------------------------------------------------------------------
    def RLSMethod(self):
        Y = np.array([0, 0, 0, 0])
        U = np.array([0, 0, 0, 0])
        PCovariant1 = 1e6*np.eye(6)
        THETA1 = np.array([1, 0, 0, 0, 0, 0])
        THETA_BUFF = np.column_stack([[0, 0, 0, 0, 0, 0]])
        a = None
        b = None

        i = 0
        for x in range(len(self.OUT)):
            Y = np.append(Y, self.OUT[i])
            U = np.append(U, self.IN[i])
            [a, b] = self.RLSMethodCore(U, Y, PCovariant1, THETA1)
            THETA1 = a
            PCovariant1 = b
            THETA_BUFF = np.column_stack([THETA_BUFF, THETA1])
            i += 1

        self.THETA_BUFF = THETA_BUFF[:, 1:]
        self.THETA_R = THETA1

    #----------------------------------------------------------------------
    def RLSMethodCore(self, U, Y, P0, TH0):
        phi = np.array([-Y[-2], -Y[-3], -Y[-4], U[-2], U[-3], U[-4]]);
        eps = Y[-1] - (phi @ TH0);
        K = (P0 @ phi) / (1 + (phi @ P0 @ phi))
        temp = phi @ P0
        PCovariant = P0 - (K.reshape(len(K), 1) @ temp.reshape(1, len(temp)))
        THETA = TH0 + K*eps

        return [THETA, PCovariant]

    #----------------------------------------------------------------------
    def CreateModelsAndOutputs(self):
        #LS
        self.LSmodel = signal.TransferFunction([self.THETA[2], self.THETA[3]], [1, self.THETA[0], self.THETA[1]], dt = self.SAMPLE_PER)
        self.T1, self.LSOUT = signal.dlsim(self.LSmodel, self.IN, t = self.TIME)
        #RLS
        self.RLSmodel = signal.TransferFunction([self.THETA_R[3], self.THETA_R[4], self.THETA_R[5]], [1, self.THETA_R[0], self.THETA_R[1], self.THETA_R[2]], dt = self.SAMPLE_PER)
        self.T2, self.RLSOUT = signal.dlsim(self.RLSmodel, self.IN, t = self.TIME)

    #----------------------------------------------------------------------
    def PrintResults(self):
        print('\nParameters from method of Least Squares:\n\na1 = '
              + str(self.THETA[0]) + '\na2 = ' + str(self.THETA[1]) + '\nb1 = '
              + str(self.THETA[2]) + '\nb2 = ' + str(self.THETA[3]) + '\n')
        print('Discrete model1 from this method')
        print(self.LSmodel)
        print('Parameters from recursive method of Least Squares:'
              '\n\na1 = ' + str(self.THETA_R[0]) + '\na2 = ' + str(self.THETA_R[1])
              + '\na3 = ' + str(self.THETA_R[2]) + '\nb1 = ' + str(self.THETA_R[3])
              + '\nb2 = ' + str(self.THETA_R[4]) + '\nb3 = ' + str(self.THETA_R[5])
              + '\n')
        print('Discrete model2 from this method')
        print(self.RLSmodel)

    #----------------------------------------------------------------------
    def PlotResults(self):
        self.panelPlot3.figure.clf()
        axes = self.panelPlot3.figure.gca()
        axes.plot(self.TIME, self.IN, 'b', linewidth = 2)
        axes.plot(self.TIME, self.OUT, 'r', linewidth = 2)
        axes.plot(self.TIME, self.LSOUT, 'm', linewidth = 1, linestyle='dashed')
        axes.plot( self.TIME, self.RLSOUT, 'k', linewidth = 1, linestyle='dashed')
        axes.grid(True)
        axes.set_xlabel('t [s]')
        axes.set_ylabel('u(t), y(t)')
        axes.set_title('Input and output signals from identified system')
        axes.legend(['input u(t)', 'output y(t)', 'model1 Least Sq.', 'model2 recursive LS'])

        self.panelPlot3.DrawSelf()

    #----------------------------------------------------------------------
    def OnCheckboxEvent(self, event):
        self.panelPlot3.figure.clf()
        axes = self.panelPlot3.figure.gca()

        flag = 0
        legendList = []

        if (self.cb31.GetValue()):
            flag += 1
            axes.plot(self.TIME, self.IN, 'b', linewidth = 2)
            legendList.append('input u(t)')
        if (self.cb32.GetValue()):
            flag += 1
            axes.plot(self.TIME, self.OUT, 'r', linewidth = 2)
            legendList.append('output y(t)')
        if (self.cb33.GetValue()):
            flag += 1
            axes.plot(self.TIME, self.LSOUT, 'm', linewidth = 1, linestyle='dashed')
            legendList.append('model1 Least Sq.')
        if (self.cb34.GetValue()):
            flag += 1
            axes.plot( self.TIME, self.RLSOUT, 'k', linewidth = 1, linestyle='dashed')
            legendList.append('model2 recursive LS')

        if (flag > 0):
            axes.grid(True)
            axes.set_xlabel('t [s]')
            axes.set_ylabel('u(t), y(t)')
            axes.set_title('Input and output signals from identified system')
            axes.legend(legendList)




        self.panelPlot3.DrawSelf()


# END class Identification ------------------------------------------------

# MAIN
if __name__ == "__main__":
    app = wx.App(False)
    mainFrame = Identification(None)
    app.MainLoop()
