%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   Identification from step input   %
%   date: 24/5/2019                  %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

close all, clear all, clc;

disp('++++++++++++++++++++++++++++++++++++++++++++++');
disp('|       Identification from step input       |');
disp('++++++++++++++++++++++++++++++++++++++++++++++');  

% DATA PREPARATION
stepValue = 1;
tRange = 40;
tSample = 0.01;     
t = [0:tSample:tRange]; 
u = ones(1, length(t)) * stepValue; u(1) = 0;

% DEFINITION OF SYSTEM
fprintf('\nSystem for identification:\n\n');
K = 5;
s = tf('s');
sys = K/((3*s+1)*(2.5*s+1)*(2.5*s+1)) 

% SIMULATING OF APPLICATION OF STEP INTO A SYSTEM
[y, ~] = lsim(sys, u, t);
a = -0.5; b = 0.5;

% GETTING THE INFLEX POINT
der1 = diff(y) / tSample;   % 1st derivation
der1 = [0 ; der1];          % to have same length as y
[~, indexDer1] = max(abs(der1));
slopeOfTangent = der1(indexDer1);
yi = y(indexDer1); ti = t(indexDer1);

% GETTING Ks, Tu, Tn, Tau
q = yi - slopeOfTangent*ti;	% Tanget: y(t) = kt + q
Ks = y(end) / stepValue;
Tu = -q / slopeOfTangent;
Tn = ((y(end) - q) / slopeOfTangent) - Tu;
Tau = Tu/Tn;
tTanget = [0:tSample:(Tn+Tu)];

% IDENTIFICATION
fprintf('\nIdentified system:\n');
if (Tau < 0.1)  % model of 2nd order different T constants
  y072 = 0.72*y(end);
  t1Index = find(abs(y) > abs(y072), 1);
  sum = t(t1Index) / 1.2564;
  t2 = (0.3574*sum);  
  y2Index = find(t > t2, 1);  
  Ynorm = y / Ks; %normalization to get y2 for table
  valueToInterpolate = abs(Ynorm(y2Index));
  
  y2 = [0.3 0.28 0.26 0.24 0.22 0.2 0.18 0.1611];
  Tau2 = [0.0 0.0435 0.0837 0.128 0.1838 0.2639 0.4031 1.0]; 
  interpolatedOutput = interp1(y2, Tau2, ...
                      valueToInterpolate,'linear', 'extrap'); 
                       
                           
  % TIME CONSTANTS  sum = T1+T2, interpolatedOutput = T2/T1
  % T1 = sum/(1+interpolatedOutput), T2 = sum - T1
  T1 = sum / (1 + interpolatedOutput); 
  T2 = sum - T1; 
  G = tf(abs(Ks), [T1*T2, T1+T2, 1])
  
  fprintf('\n');
  disp('Identified model is of 2nd order with parameters:');
  fprintf('\n');
  disp(['T1 = ' num2str(T1)]); disp(['T2 = ' num2str(T2)]);
  disp(['Ks = ' num2str(abs(Ks))]);  
else    % model of n-th order, same Time constants
  Tau3 = [0.104 0.218 0.319 0.410 0.493 0.570 ...
          0.642 0.709 0.773];
  order = [2 3 4 5 6 7 8 9 10];
  n = round(interp1(Tau3, order, Tau,'linear', 'extrap'));
  Tm = ti/(n - 1);
  s = tf('s');
  G = abs(Ks)/((Tm*s+1)^n)
  
  fprintf('\n');
  disp(['Identified model is of order ' num2str(n) ':']);
  fprintf('\n');
  disp(['T = ' num2str(Tm)]); disp(['n = ' num2str(n)]);
  disp(['Ks = ' num2str(abs(Ks))]);
end

% SIMULATION OF INPUT TO COMPUTED MODEL
[yModel, ~] = lsim(G, u, t);

% PLOTTING
plot(t, u, 'g-.', 'LineWidth', 2);    
hold on;
plot(t, y, 'r-.', 'LineWidth', 2);
plot(t, yModel, 'b', 'LineWidth', 1);
plot(tTanget, slopeOfTangent*tTanget + q, 'k');
plot(ti, yi, 'kO', 'LineWidth', 2)
plot(Tu, 0, 'k+', Tn+Tu, 0, 'k+', 'LineWidth', 2);
text(ti, yi, '   \leftarrow Inflex point'); 
text(Tu, 0, '   Tu'); text(Tn+Tu, 0, '   Tu+Tn');
hold off; grid on;
xlim([-0.5 tRange+0.5]); 
if (y(end) > 0)
  ylim([-y(end)*0.05 y(end)*1.05]);
else
  ylim([y(end)*1.05 -y(end)*0.05]);
end
legend('input', 'system output', 'model output', ...
       'Tangent', 'Inflex point');           
title('\bf \fontsize{12} Identification method of sys. from step ipnut.'); 
xlabel('\bf t [s]'); ylabel('\bf u(t), y(t)');

fprintf('\n');
disp('++++++++++++++++++++++++++++++++++++++++++++++');
disp('|                     END                    |');
disp('++++++++++++++++++++++++++++++++++++++++++++++');
